# -*- coding: utf-8 -*-
import time

from django import forms
from django.apps import apps
from django.conf import settings
from django.utils.translation import gettext_lazy as _

from django_otp.oath import TOTP


TOTP_SESSION_KEY = 'kleides-mfa-totp-key'


class BaseVerifyForm(forms.Form):
    def __init__(
            self, device, plugin, request, unverified_user, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.unverified_user = unverified_user
        self.plugin = plugin
        self.device = device
        self.request = request

    def get_user(self):
        if self.is_valid():
            return self.unverified_user
        return None

    def get_device(self):
        if self.is_valid():
            return self.device
        return None


class DeviceVerifyForm(BaseVerifyForm):
    otp_token = forms.CharField(label=_('Token'))

    error_messages = {
        'invalid': _('The token is not valid for this device.'),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['otp_token'].widget.attrs.update({
            'autocomplete': 'off', 'autofocus': 'autofocus'})

    def clean(self):
        cleaned_data = super().clean()
        token = cleaned_data.get('otp_token')
        if not token:  # token is a required field.
            return cleaned_data

        # Note that tokens can become invalid once verified.
        if not self.device.verify_token(token):
            raise forms.ValidationError(self.error_messages['invalid'])
        return cleaned_data


class BaseDeviceForm(forms.ModelForm):
    def __init__(self, plugin, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.plugin = plugin
        self.request = request
        if 'name' in self.fields:
            self.fields['name'].initial = self.plugin.name
            self.fields['name'].required = False
            self.fields['name'].widget.attrs['placeholder'] = _(
                'Friendly name')

    def clean_name(self):
        if self.cleaned_data.get('name'):
            return self.cleaned_data['name']
        return self.plugin.name


class DeviceCreateForm(BaseDeviceForm):
    error_messages = {
        'invalid': _('Unable to validate the token with the device.'),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # WARNING: The form associates the user with the potential new device
        # instance to allow device functions dependent on the user to be used.
        # The plugin author must take care to only use functions that will not
        # save the instance as this may circumvent the verification process.
        self.instance.user = self.request.user
        self.instance.confirmed = False

    def _post_clean(self):
        super()._post_clean()
        if self.instance.pk is not None:
            raise ValueError(
                'Plugin {} saved the device before completing verification. '
                'The device will be unconfirmed and unusable. '
                'Please review {} {!r}.'.format(
                    self.plugin, self.instance.persistent_id, self.instance))
        self.instance.confirmed = True


class DeviceUpdateForm(BaseDeviceForm):
    pass


class DeviceDeleteForm(BaseDeviceForm):
    pass


if apps.is_installed('django_otp.plugins.otp_totp'):  # noqa: E501 C901
    from django_otp.plugins.otp_totp.models import TOTPDevice

    class TOTPDeviceCreateForm(DeviceCreateForm):
        otp_token = forms.CharField(label=_('Token'))

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.fields['otp_token'].widget.attrs.update({
                'autocomplete': 'off', 'autofocus': 'autofocus'})
            # Store the TOTP key in the session and rotate the key on unbound
            # forms.
            if self.is_bound and TOTP_SESSION_KEY in self.request.session:
                self.instance.key = self.request.session[TOTP_SESSION_KEY]
            else:
                self.request.session[TOTP_SESSION_KEY] = self.instance.key
            # If alternate TOTPDevice settings are needed set them here.
            # self.instance.digits = 8  # increase token length.
            # self.instance.tolerance = 2  # for users that are veeeeery slow.

        def clean(self):
            cleaned_data = super().clean()
            try:
                token = int(cleaned_data.get('otp_token'))
            except (TypeError, ValueError):
                verified = False
            else:
                # django-otp setting.
                OTP_TOTP_SYNC = getattr(settings, 'OTP_TOTP_SYNC', True)
                # Device verification using the current instance.
                totp = TOTP(
                    self.instance.bin_key, self.instance.step,
                    self.instance.t0, self.instance.digits,
                    self.instance.drift)
                totp.time = time.time()

                verified = totp.verify(
                    token, self.instance.tolerance, self.instance.last_t)
                if verified:
                    # Device is verified, update attributes and prepare the
                    # instance to be saved.
                    self.instance.last_t = totp.t()
                    if OTP_TOTP_SYNC:
                        self.instance.drift = totp.drift
            if not verified:
                raise forms.ValidationError(self.error_messages['invalid'])
            try:
                return cleaned_data
            finally:
                if TOTP_SESSION_KEY in self.request.session:  # noqa: E501 pragma: no cover
                    del self.request.session[TOTP_SESSION_KEY]

        class Meta:
            model = TOTPDevice
            fields = ('name', 'otp_token',)


if apps.is_installed('django_otp.plugins.otp_static'):
    from django_otp.plugins.otp_static.models import StaticDevice, StaticToken

    class RecoveryDeviceForm(DeviceUpdateForm):
        token_amount = 10

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            try:
                self.instance = self.request.user.staticdevice_set.get()
            except self.request.user.staticdevice_set.model.DoesNotExist:
                pass

        def save(self, *args, **kwargs):
            instance, created = (
                self.request.user.staticdevice_set.get_or_create(
                    defaults={'name': self.plugin.name}))
            instance.token_set.all().delete()
            for i in range(self.token_amount):
                instance.token_set.create(token=StaticToken.random_token())
            return instance

        class Meta:
            model = StaticDevice
            fields = ()


if apps.is_installed('otp_yubikey'):  # pragma: no branch
    from otp_yubikey.models import RemoteYubikeyDevice, ValidationService

    class YubikeyDeviceCreateForm(DeviceCreateForm):
        service = forms.ModelChoiceField(
            label=_('Service'), queryset=ValidationService.objects.all(),
            empty_label=None)
        otp_token = forms.CharField(label=_('Token'))

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.fields['otp_token'].widget.attrs.update({
                'autocomplete': 'off', 'autofocus': 'autofocus'})
            try:
                self.fields['service'].initial = (
                    self.fields['service'].queryset.get().pk)
                self.fields['service'].widget = forms.HiddenInput()
            except ValidationService.MultipleObjectsReturned:
                # Multiple is good, none is bad.
                pass

        def clean(self):
            cleaned_data = super().clean()
            token = self.cleaned_data.get('otp_token')
            service = self.cleaned_data.get('service')
            verified = False
            if token and service:
                self.instance.service = service
                self.instance.public_id = token[:-32]
                verified = self.instance.verify_token(token)
            if not verified:
                raise forms.ValidationError(self.error_messages['invalid'])
            return cleaned_data

        class Meta:
            model = RemoteYubikeyDevice
            fields = ('service', 'name', 'otp_token',)



from datetime import timedelta

from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import PermissionDenied
from django.http import HttpResponse
from django.test import RequestFactory, TestCase
from django.test.utils import override_settings
from django.utils import timezone

from django_otp import DEVICE_ID_SESSION_KEY

from kleides_mfa.views.mixins import VERIFIED_SESSION_KEY
from kleides_mfa.decorators import (
    create_decorator, single_factor_required, multi_factor_required,
    recent_multi_factor_required, setup_or_mfa_required,
    setup_or_recent_mfa_required)

from .factories import UserFactory


class KleidesMfaDecoratorTestCase(TestCase):
    def setUp(self):
        self.rfactory = RequestFactory()

    def request(self, user):
        request = self.rfactory.get('/view_test/')
        request.user = user
        request.session = {}
        return request

    def mfa_request(self, user, verified_on=None):
        device = user.totpdevice_set.get_or_create(name='test')[0]
        # Force the user to login with 2 step authentication.
        # django-otp should do this if the otp_device is set.
        user.otp_device = device

        request = self.request(user)

        request.session[DEVICE_ID_SESSION_KEY] = device.persistent_id
        if verified_on is None:
            verified_on = timezone.now()
        request.session[VERIFIED_SESSION_KEY] = verified_on.isoformat()
        return request

    def test_create_decorator(self):
        with self.assertRaises(ValueError):
            create_decorator(lambda r: r.user.is_verified)

    def test_args(self):
        @single_factor_required()
        def view_test(request):
            return HttpResponse('view_test')

        request = self.request(AnonymousUser())
        response = view_test(request)
        self.assertEqual(response.status_code, 302)

    def test_no_args(self):
        @single_factor_required
        def view_test(request):
            return HttpResponse('view_test')

        request = self.request(AnonymousUser())
        response = view_test(request)
        self.assertEqual(response.status_code, 302)

    def test_single_factor_required(self):
        @single_factor_required(raise_exception=True)
        def view_test(request):
            return HttpResponse('view_test')

        with self.assertRaises(PermissionDenied):
            request = self.request(AnonymousUser())
            view_test(request)

        request = self.request(UserFactory())
        response = view_test(request)
        self.assertEqual(response.status_code, 200)

    def test_multi_factor_required(self):
        @multi_factor_required(raise_exception=True)
        def view_test(request):
            return HttpResponse('view_test')

        with self.assertRaises(PermissionDenied):
            request = self.request(AnonymousUser())
            view_test(request)

        with self.assertRaises(PermissionDenied):
            request = self.request(UserFactory())
            view_test(request)

        request = self.mfa_request(UserFactory())
        response = view_test(request)
        self.assertEqual(response.status_code, 200)

    @override_settings(KLEIDES_MFA_VERIFIED_TIMEOUT=60)
    def test_recent_multi_factor_required(self):
        @recent_multi_factor_required(raise_exception=True)
        def view_test(request):
            return HttpResponse('view_test')

        with self.assertRaises(PermissionDenied):
            request = self.request(AnonymousUser())
            view_test(request)

        with self.assertRaises(PermissionDenied):
            request = self.request(UserFactory())
            view_test(request)

        with self.assertRaises(PermissionDenied):
            request = self.mfa_request(
                UserFactory(),
                verified_on=timezone.now() - timedelta(seconds=90))
            view_test(request)

        request = self.mfa_request(UserFactory())
        response = view_test(request)
        self.assertEqual(response.status_code, 200)

    def test_setup_or_mfa_required(self):
        @setup_or_mfa_required(raise_exception=True)
        def view_test(request):
            return HttpResponse('view_test')

        with self.assertRaises(PermissionDenied):
            request = self.request(AnonymousUser())
            view_test(request)

        with self.assertRaises(PermissionDenied):
            request = self.request(UserFactory())
            request.user.totpdevice_set.get_or_create(name='test')
            view_test(request)

        request = self.request(UserFactory())
        response = view_test(request)
        self.assertEqual(response.status_code, 200)

        request = self.mfa_request(UserFactory())
        response = view_test(request)
        self.assertEqual(response.status_code, 200)

    @override_settings(KLEIDES_MFA_VERIFIED_TIMEOUT=60)
    def test_setup_or_recent_mfa_required(self):
        @setup_or_recent_mfa_required(raise_exception=True)
        def view_test(request):
            return HttpResponse('view_test')

        with self.assertRaises(PermissionDenied):
            request = self.request(AnonymousUser())
            view_test(request)

        with self.assertRaises(PermissionDenied):
            request = self.request(UserFactory())
            request.user.totpdevice_set.get_or_create(name='test')
            view_test(request)

        with self.assertRaises(PermissionDenied):
            request = self.mfa_request(
                UserFactory(),
                verified_on=timezone.now() - timedelta(seconds=90))
            view_test(request)

        request = self.request(UserFactory())
        response = view_test(request)
        self.assertEqual(response.status_code, 200)

        request = self.mfa_request(UserFactory())
        response = view_test(request)
        self.assertEqual(response.status_code, 200)



from functools import wraps
from urllib.parse import urlparse

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.core.exceptions import PermissionDenied
from django.shortcuts import resolve_url
from django.utils.translation import gettext_lazy as _

from .views.mixins import is_recently_verified, is_user_in_setup


def user_passes_test(
    test_func, login_url=None, raise_exception=False,
    redirect_field_name=REDIRECT_FIELD_NAME
):
    '''
    Decorator for views that checks that the user passes the given test,
    redirecting to the log-in page if necessary. The test should be a callable
    that takes the request object and returns True if the user passes.
    If raise_exception is True a PermissionDenied exception is raised instead
    of redicting to the login page.
    '''

    def decorator(view_func):
        @wraps(view_func)
        def _wrapper_view(request, *args, **kwargs):
            if test_func(request):
                return view_func(request, *args, **kwargs)

            if raise_exception:
                raise PermissionDenied

            if request.user.is_verified:
                messages.info(
                    request,
                    _('We need to confirm your identity, please login again.'))

            path = request.build_absolute_uri()
            resolved_login_url = resolve_url(login_url or settings.LOGIN_URL)
            # If the login url is the same scheme and net location then just
            # use the path as the "next" url.
            login_scheme, login_netloc = urlparse(resolved_login_url)[:2]
            current_scheme, current_netloc = urlparse(path)[:2]
            if (not login_scheme or login_scheme == current_scheme) and (
                not login_netloc or login_netloc == current_netloc
            ):
                path = request.get_full_path()
            from django.contrib.auth.views import redirect_to_login

            return redirect_to_login(
                path, resolved_login_url, redirect_field_name)

        return _wrapper_view

    return decorator


def create_decorator(
    function=None, test_func=None, login_url=None, raise_exception=False,
    redirect_field_name=REDIRECT_FIELD_NAME
):
    if not callable(test_func):
        raise ValueError('test_func must be a callable that accepts a request')

    actual_decorator = user_passes_test(
        test_func,
        login_url=login_url, raise_exception=raise_exception,
        redirect_field_name=redirect_field_name,
    )

    if function:
        return actual_decorator(function)

    return actual_decorator


def single_factor_required(*args, **kwargs):
    '''
    Decorator for views that only required a single authentication factor.
    '''
    return create_decorator(
        *args,
        test_func=lambda r: r.user.is_single_factor_authenticated,
        **kwargs)


def multi_factor_required(*args, **kwargs):
    '''
    Decorator for views that require multi factor authentication.
    '''
    return create_decorator(
        *args,
        test_func=lambda r: r.user.is_verified,
        **kwargs)


def recent_multi_factor_required(*args, **kwargs):
    '''
    Decorator for views that require recent multi factor authentication.
    '''
    return create_decorator(*args, test_func=is_recently_verified, **kwargs)


def setup_or_mfa_required(*args, **kwargs):
    '''
    Decorator for views that require multi factor authentication or
    with single factor and is still in the process of account setup.
    '''
    return create_decorator(
        *args,
        test_func=lambda r: r.user.is_verified or is_user_in_setup(r),
        **kwargs)


def setup_or_recent_mfa_required(*args, **kwargs):
    '''
    Decorator for views that require recent multi factor authentication or
    with single factor and is still in the process of account setup.
    '''
    return create_decorator(
        *args,
        test_func=lambda r: is_recently_verified(r) or is_user_in_setup(r),
        **kwargs)



'''
These are the available settings, accessed through
``kleides_mfa.conf.app_settings``. All attributes prefixed ``KLEIDES_MFA_*``
can be overridden from your Django project's settings module by defining a
setting with the same name.

For instance, to enable mfa on the the admin interface:

.. code-block:: python

    KLEIDES_MFA_PATCH_ADMIN = True
'''
from __future__ import annotations

from dataclasses import dataclass
from typing import Any
import warnings

from django.conf import settings as django_settings

# All attributes accessed with this prefix are possible to overwrite
# through django.conf.settings.
settings_prefix = "KLEIDES_MFA_"


@dataclass(frozen=True)
class AppSettings:
    '''Access this instance as ``kleides_mfa.conf.app_settings``.'''

    # The sorting order of the plugins on the mfa listing using the
    # plugin.slug. # These are known plugins in order of security.
    # hardware tokens > software tokens > backup codes.
    KLEIDES_MFA_PLUGIN_PRIORITY: list[str] | tuple[str] = (
        'u2f', 'yubikey', 'totp', 'recovery-code',
    )

    # Patch the AdminSite class and default admin site instance to require
    # 2 step authentication.
    KLEIDES_MFA_PATCH_ADMIN: bool = True

    # The default url to redirect to after login when the user has no 2 step
    # authentication devices active on the account.
    KLEIDES_MFA_SINGLE_FACTOR_URL: str = 'kleides_mfa:index'

    # Amount of seconds before a user is required to verify his authenticate
    # again to pass the recently authenticated requirement. This is can be used
    # to protect views that modify authentication settings.
    KLEIDES_MFA_VERIFIED_TIMEOUT: int | None = 900

    # Update the verification time on every request that passes the
    # recently_verified test to act as recent activity tracking. This prevents
    # a forced verified timeout while the user is actively using the account.
    KLEIDES_MFA_VERIFIED_UPDATE: bool = True

    def __getattribute__(self, name: str) -> Any:
        '''
        Check if a Django project settings should override the app default.

        Only allow settings from the settings prefix in order to avoid
        returning any random properties of the django settings we the settings
        prefix.
        '''
        if name == 'KLEIDES_MFA_PATCH_ADMIN':
            value = getattr(django_settings, 'KLEIDES_MFA_PATCH_ADMIN', True)
            if value:
                warnings.warn(
                    'The KLEIDES_MFA_PATCH_ADMIN=True setting is deprecated. '
                    'Use the default_site="kleides_mfa.admin.KleidesMfaAdmin" '
                    'attribute on your project AppConfig and remove the '
                    'django.contrib.admin app. '
                    'Then set KLEIDES_MFA_PATCH_ADMIN to False. '
                    'https://docs.djangoproject.com/en/5.0/ref/contrib/admin/'
                    '#overriding-the-default-admin-site', DeprecationWarning)
            return value

        if name.startswith(settings_prefix) and hasattr(django_settings, name):
            return getattr(django_settings, name)

        return super().__getattribute__(name)


app_settings = AppSettings()
